/*
 * --------------------------------[ title ]-----------------------------------
 *                                                       (C)1998 RedLemon Tools
 * name:        xmodpointer.c
 * author:      Lyon Lemmens
 * description: Modify pointer button settings for XInput devices
 *
 * history:
 * ~~~~~~~~
 * when          who             what
 * ------------  --------------  ------------------------
 * Tue 17.03.98  Lyon Lemmens    initial version
 *
 * ----------------------------------------------------------------------------
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xproto.h>
#include <X11/extensions/XInput.h>

/*
 * ============================================================================
 * === macros =================================================================
 * ============================================================================
 */
    /*
     * --- Version number -----------------------------------------------------
     */
#define VERSION     "0.1"

    /*
     * --- Extract the name from a XDeviceInfo struct -------------------------
     */
#define DeviceName(dev_)    ((dev_)->name? (dev_)->name : "<noname>")

    /*
     * --- Debugging macro ----------------------------------------------------
     */
#ifdef DEBUG
#   define  BURP(code_)    {code_;}
#else
#   define  BURP(code_)
#endif

    /*
     * --- Maximum size of the buttonmap --------------------------------------
     */
#define MAPSIZE     (256)

    /*
     * --- MIN/MAX ------------------------------------------------------------
     */
#define MIN(a_,b_)  (((a_) < (b_))? (a_) : (b_))
#define MAX(a_,b_)  (((a_) > (b_))? (a_) : (b_))

/*
 * ============================================================================
 * === globals ================================================================
 * ============================================================================
 */
    /*
     * --- New buttonmap to install (if any) ----------------------------------
     */
static unsigned char install_map[MAPSIZE];
static int           install_nr = 0;

    /*
     * --- Pointer name to change (if any) ------------------------------------
     */
static char *pointer_name;

    /*
     * --- noisiness of operation ---------------------------------------------
     */
static int noise = 1;

/*
 * ============================================================================
 * === prototypes =============================================================
 * ============================================================================
 */
static int  find_xinput_extension (Display*);
static int  find_pointer          (Display*);
static void show_device_buttons   (Display*, XDeviceInfo*);
static void set_device_buttons    (Display*, XDeviceInfo*);
static void usage                 (void);

/*
 * ============================================================================
 * === implementation =========================================================
 * ============================================================================
 */
    /*
     * --- Case insensitive string compare ------------------------------------
     */
static int
str_icmp
    (char *str1, char *str2)
{
    /*
     * 1. A NULL string is allways lexically less than a non-NULL string.
     * 2. Two NULL strings are equal.
     * 3. Two non-NULL strings must be compared.
     */
    if (str1 == NULL)
            return (str2 == NULL)? 0 : -1;

    /*
     * at this point str1 can't be NULL
     */
    if (str2 == NULL)
            return 1;

    /*
     * empty string is always the smallest
     */
	if (*str1 == 0)
            return (*str2 != 0);

    /*
     * compare strings until strings stop or strings are unequal
     */
	while (tolower (*str1) == tolower (*str2))
	{
		if (*str1 == '\0')
                return (0);
		str1++;
        str2++;
	}

    /*
     * return difference
     */
	return (tolower (*str1) - tolower (*str2));
}

    /*
     * --- Look for the XInput extension --------------------------------------
     */
static int
find_xinput_extension
    (Display *dpy)
{
    char  **extension;
    int   ii, count;

    extension = XListExtensions (dpy, &count);

    for (ii = 0; ii < count; ii++)
    {
        BURP (fprintf (stderr, "found extension %s\n", extension[ii]));

        if (strcmp (extension[ii], "XInputExtension") == 0)
                break;
    }
    XFreeExtensionList (extension);

    return (ii != count);
}

    /*
     * --- Look for the requested pointer -------------------------------------
     */
static int
find_pointer
    (Display *dpy)
{
    XDeviceInfo   *device_list, *device;
    int            ii, count, status = 0;

    device_list = XListInputDevices (dpy, &count);
    for (ii = 0, device = device_list; ii < count; ii++, device++)
    {
        BURP (fprintf (stderr, "found device %s\n", DeviceName (device)));

        /*
         * cannot set buttons on core devices
         */
        if (device->use == IsXExtensionDevice)
        {
            /*
             * program called without parms: display all devices and
             * their button maps
             */
            if (pointer_name == NULL)
            {
                show_device_buttons (dpy, device);
                status = 1;
            }

            /*
             * specific device requested
             */
            else if (str_icmp (pointer_name, DeviceName (device)) == 0)
            {
                /*
                 * install a new button map?
                 */
                if (install_nr != 0)
                {
                    set_device_buttons  (dpy, device);
                    show_device_buttons (dpy, device);
                    status = 1;
                    break;
                }

                /*
                 * No -> show the current map for this device only
                 */
                else
                {
                    show_device_buttons (dpy, device);
                    status = 1;
                    break;
                }
            }
        }
    }

    /*
     * clean up and return the status (true|false)
     */
    XFreeDeviceList (device_list);
    return status;
}

    /*
     * --- Show the buttonmapping for a device --------------------------------
     */
static void
show_device_buttons
    (Display     *dpy,
     XDeviceInfo *device)
{
    XDevice *devptr;

    BURP (fprintf (stderr, "opening device %s\n", DeviceName (device)));

    devptr = XOpenDevice (dpy, device->id);
    if (devptr != NULL)
    {
        unsigned char map[MAPSIZE];
        int           ii, count;

        count = XGetDeviceButtonMapping (dpy, devptr, map, MAPSIZE);
        if (noise)
        {
            if (count > 0)
            {
                printf ("Button mapping for device %s:\n", DeviceName (device));
                for (ii = 0; ii < count; ii++)
                {
                    printf ("  %2d", map[ii]);
                }
                printf ("\n\n");
            }
            else
            {
                printf ("Device %s has no button mapping\n",
                        DeviceName (device));
            }
        }
    }
    else
    {
        fprintf (stderr, "xmodpointer: Error opening device %s\n",
                    DeviceName (device));
    }
    XCloseDevice (dpy, devptr);
}

    /*
     * --- Set the buttonmapping for a device ---------------------------------
     */
static void
set_device_buttons
    (Display     *dpy,
     XDeviceInfo *device)
{
    XDevice *devptr;

    BURP (fprintf (stderr, "opening device %s\n", DeviceName (device)));

    devptr = XOpenDevice (dpy, device->id);
    if (devptr != NULL)
    {
        unsigned char map[MAPSIZE];
        int           count, result;

        /*
         * don't map more than the allowed number of keys
         */
        count = XGetDeviceButtonMapping (dpy, devptr, map, MAPSIZE);
        count = MIN (count, install_nr);
        result = XSetDeviceButtonMapping (dpy, devptr, install_map, count);
        if (result != MappingSuccess)
        {
            fprintf (stderr, "Button remap for device %s failed\n", 
                     DeviceName (device));
        }
    }
    else
    {
        fprintf (stderr, "xmodpointer: Error opening device %s\n",
                    DeviceName (device));
    }
    XCloseDevice (dpy, devptr);
}
    
    /*
     * --- Print usage message ------------------------------------------------
     */
static void
usage
    (void)
{
    printf ("xmodpointer version %s -- by Lyon Lemmens\n\n"
            "usage:\n"
            "   xmodpointer [-h] | [device [buttonmap]]\n"
            "\n"
            "where:\n"
            "   -h          Gives this message.\n"
            "   -q          Quiet operation.\n"
            "   device      Is the name of the XInput device to change.\n"
            "   buttonmap   Is the new buttonmap for this device.\n"
            "\n"
            "If no device is given, xmodpointer prints the buttonmaps\n"
            "of all devices it encounters.\n"
            "\n"
            "If no buttonmap is given, xmodpointer prints the buttonmap\n"
            "of the named device only.\n"
            "\n", VERSION
            );
}

    /*
     * --- MAIN ---------------------------------------------------------------
     */
int
main
    (int   argc,
     char *argv[])
{
    Display *dpy;

    /*
     * parse command line
     */
    if (argc >= 2)
    { 
        int ii = 1;

        if (strcmp (argv[ii], "-h") == 0 || strcmp (argv[ii], "-?") == 0)
        {
            usage ();
            return (0);
        }

        if (strcmp (argv[ii], "-q") == 0)
        {
            noise = 0;
            ii++;
        }

        pointer_name = argv[ii];
        for (ii++,        install_nr = 0;
             ii < argc && install_nr < MAPSIZE;
             ii++,        install_nr++)
        {
            install_map[install_nr] = (unsigned char)atoi (argv[ii]);
        }
    }

    /*
     * Open the connection to the display
     */
    dpy = XOpenDisplay (NULL);
    if (dpy == NULL) {
        fprintf (stderr, "xmodpointer: unable to connect to X Server\n");
        return 1;
    }

    BURP (fprintf (stderr, "connected to %s\n", XDisplayString (dpy)))

    /*
     * look for the XInput extension
     */
    if (!find_xinput_extension (dpy))
    {
        fprintf (stderr,
                 "xmodpointer: could not find a XInput Extension "
                 "on this display\n");
        return 2;
    }

    /*
     * list all pointers with their button settings
     */
    if (!find_pointer (dpy))
    {
        fprintf (stderr, "xmodpointer: couldn't find a pointer ");
        if (pointer_name != NULL)
                fprintf (stderr, "\"%s\" ", pointer_name);
        fprintf (stderr, "on this display\n");
        return 3;
    }

    return 0;
}
